import ast


def parse_kwarg_keys(source, keywords='kwargs', with_vals=False):
    r"""
    Parses the source code to find keys used by the `**kwargs` keywords
    dictionary variable. if `with_vals` is True, we also attempt to infer the
    default values.

    Args:
        source (str):

    Returns:
        list: kwarg_keys

    SeeAlso:
        argparse_funckw
        recursive_parse_kwargs
        parse_kwarg_keys
        parse_func_kwarg_keys
        get_func_kwargs

    Example:
        >>> import ubelt as ub
        >>> source = ub.codeblock(
        >>>    '''
        >>>    x = 'hidden_x'
        >>>    y = 3 # hidden val
        >>>    kwargs.get(x, y)
        >>>    kwargs.get('foo', None)
        >>>    kwargs.pop('bar', 3)
        >>>    kwargs.pop('str', '3fd')
        >>>    kwargs.pop('str', '3f"d')
        >>>    "kwargs.get('baz', None)"
        >>>    kwargs['foo2']
        >>>    #kwargs.get('biz', None)"
        >>>    kwargs['bloop']
        >>>    x = 'bop' in kwargs
        >>>    ''')
        >>> print('source = %s\n' % (source,))
        >>> with_vals = True
        >>> kwarg_items = parse_kwarg_keys(source, with_vals=with_vals)
        >>> result = ('kwarg_items = %s' % (ub.repr2(kwarg_items, nl=1),))
        >>> kwarg_keys = [item[0] for item in kwarg_items]
        ...
        >>> print(result)
        kwarg_items = [
            ('foo', None),
            ('bar', 3),
            ('str', '3fd'),
            ('str', '3f"d'),
            ('foo2', None),
            ('bloop', None),
        ]
        >>> assert 'baz' not in kwarg_keys
        >>> assert 'foo' in kwarg_keys
        >>> assert 'bloop' in kwarg_keys
        >>> assert 'bop' not in kwarg_keys
    """
    pt = ast.parse(source)
    kwargs_items = []
    debug = 1
    target_kwargs_name = keywords

    class KwargParseVisitor(ast.NodeVisitor):
        """
        A visitor to parse keyword arguments passed to functions in the AST.

        This will look for functions with kwargs, parse calls to `get` or `pop`,
        and handle constants and expressions as they are found in function signatures.
        """
        def __init__(self):
            super().__init__()
            self.const_lookup = {}
            self.first = True
            self.target_kwargs_name = target_kwargs_name
            self.kwargs_items = kwargs_items

        def visit_FunctionDef(self, node):
            if debug:
                print(f"VISIT FunctionDef node = {node!r}")

            # Get the keyword argument (if any)
            kwarg_name = node.args.kwarg.arg if node.args.kwarg else None

            # Record any constants defined in function definitions
            defaults_vals = node.args.defaults
            offset = len(node.args.args) - len(defaults_vals)
            default_keys = node.args.args[offset:]

            for kwname, kwval in zip(default_keys, defaults_vals):
                if isinstance(kwval, ast.NameConstant):
                    val = kwval.value
                    self.const_lookup[kwname.arg] = val

            # If target kwargs are still in scope, visit the node
            if self.first or kwarg_name != self.target_kwargs_name:
                ast.NodeVisitor.generic_visit(self, node)
                self.first = False

        def visit_Subscript(self, node):
            if debug:
                print(f"VISIT SUBSCRIPT node = {node!r}")
                print(node.value.id)
                print(node.value)
                print(node.slice)

            if isinstance(node.value, ast.Name) and node.value.id == self.target_kwargs_name:
                if isinstance(node.slice, ast.Index):
                    index = node.slice
                    key = index.value
                    if isinstance(key, (ast.Str, ast.Constant)):  # Support both for older and newer Python versions
                        print(f'key={key}')
                        key_value = key.s if isinstance(key, ast.Str) else key.value  # Extract the key's value
                        item = (key_value, None)  # Default value as None
                        self.kwargs_items.append(item)
                elif isinstance(node.slice, ast.Constant):
                    item = (node.slice.value, None)  # Default value as None
                    self.kwargs_items.append(item)

        @staticmethod
        def _eval_bool_op(val):
            val_value = None
            if isinstance(val.op, ast.Or):
                if any(isinstance(x, ast.NameConstant) and x.value is True for x in val.values):
                    val_value = True
            elif isinstance(val.op, ast.And):
                if any(isinstance(x, ast.NameConstant) and x.value is False for x in val.values):
                    val_value = False
            return val_value

        def visit_Call(self, node):
            if debug:
                print(f"VISIT Call node = {node!r}")

            if isinstance(node.func, ast.Attribute):
                try:
                    objname = node.func.value.id
                except AttributeError:
                    return
                methodname = node.func.attr

                if objname == self.target_kwargs_name and methodname in {'get', 'pop'}:
                    args = node.args
                    if len(args) == 2:
                        key, val = args
                        if isinstance(key, ast.Name):
                            pass  # TODO: lookup constant if necessary
                        elif isinstance(key, ast.Str):
                            key_value = key.s
                            val_value = self._parse_value(val)
                            item = (key_value, val_value)
                            self.kwargs_items.append(item)

            ast.NodeVisitor.generic_visit(self, node)

        def _parse_value(self, val):
            """ Helper to parse different types of AST nodes and return their values """
            if isinstance(val, ast.Str):
                return val.s
            elif isinstance(val, ast.Num):
                return val.n
            elif isinstance(val, ast.Name):
                if val.id == 'None':
                    return None
                return self.const_lookup.get(val.id, None)
            elif isinstance(val, ast.NameConstant):
                return val.value
            elif isinstance(val, ast.Call):
                return None  # You can handle Call as necessary
            elif isinstance(val, ast.BoolOp):
                return self._eval_bool_op(val)
            elif isinstance(val, ast.Dict):
                return {}  # You can handle Dict if necessary
            else:
                print(f"Warning: util_inspect doesn't know how to parse {repr(val)}")
                return None

    try:
        KwargParseVisitor().visit(pt)
    except Exception:
        raise
        pass
    if with_vals:
        return kwargs_items
    else:
        return [item[0] for item in kwargs_items]

if __name__ == '__main__':
    """
    CommandLine:
        python -m xinspect.static_kwargs all
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



import os
import warnings
import tempfile
import collections
from collections import OrderedDict


def undefined_names(fpath=None, source=None):
    """
    Use a linter to find undefined names in a Python file

    Args:
        fpath (PathLike): path to the file
        source (str): source code of file (mutually exclusive with fpath)

    Example:
        >>> import ubelt as ub
        >>> source = ub.codeblock(
        >>>     '''
        >>>     p = os.path.dirname(join('a', 'b'))
        >>>     glob.glob(p)
        >>>     ''')
        >>> sorted(undefined_names(source=source))
        ['glob', 'join', 'os']
    """
    import pyflakes.api
    import pyflakes.reporter

    if not (bool(source) ^ bool(fpath)):
        raise ValueError('Must specify exactly one fpath or source')

    class CaptureReporter(pyflakes.reporter.Reporter):
        def __init__(reporter, warningStream, errorStream):
            reporter.syntax_errors = []
            reporter.messages = []
            reporter.unexpected = []

        def unexpectedError(reporter, filename, msg):
            reporter.unexpected.append(msg)

        def syntaxError(reporter, filename, msg, lineno, offset, text):
            reporter.syntax_errors.append(msg)

        def flake(reporter, message):
            reporter.messages.append(message)

    # TODO: use code transformer to remove import * by default
    names = set()
    reporter = CaptureReporter(None, None)
    if source is not None:
        pyflakes.api.check(source, '_.py', reporter)
    else:
        pyflakes.api.checkPath(fpath, reporter)
    for msg in reporter.messages:
        if msg.__class__.__name__.endswith('UndefinedName'):
            assert len(msg.message_args) == 1
            names.add(msg.message_args[0])
    return names


class Importables:
    """
    Class that keeps track of registered known importables
    """
    def __init__(self, default=None):
        # A mapping from undefined names in a file to the appropriate line of
        # Python code that defines that name (usually an import)
        if default is None:
            default = OrderedDict()
        elif isinstance(default, Importables) or hasattr(default, 'known'):
            default = default.known
        self.known = default

    def update(self, other):
        self.known.update(other)

    def __getitem__(self, key):
        return self.known[key]

    def __iter__(self):
        return iter(self.known)

    def _use_recommended_defaults(self):
        """
        Adds a list of default values that I like.
        This should change to be much more configurable.
        """
        self._populate_common_modules()
        self._populate_common_aliases()
        self._populate_uncommon_aliases()
        self._populate_ubiquitous_stdlib_members()

    def _populate_common_modules(self):
        modules = [
            'cv2', 'glob', 'torch', 'math'
        ]
        for name in modules:
            self.known[name] = 'import {}'.format(name)

    def _populate_common_aliases(self):
        self.known.update({
            'np': 'import numpy as np',
            'pd': 'import pandas as pd',
            'mpl': 'import matplotlib as mpl',
        })

    def _populate_uncommon_aliases(self):
        self.known.update({
            'it': 'import itertools as it',
            'nn': 'from torch import nn',
            'F': 'import torch.nn.functional as F',
            'Image': 'from PIL import Image',
            'ub': 'import ubelt as ub',
            'nh': 'import netharn as nh',
        })

    def _populate_ubiquitous_stdlib_members(self):
        # Add os.path members to aliased importables
        for name in dir(os.path):
            if not name.startswith('_') and name != 'os':
                self.known[name] = 'from os.path import {}'.format(name)
        for key in collections.__all__:
            self.known[key] = 'from collections import {}'.format(key)

        # Note: not all ctypes members are ubuquitous
        # ctypes_members = [d for d in dir(ctypes) if not d.startswith('_')]
        ctypes_members = [
            'c_bool', 'c_buffer', 'c_byte', 'c_char', 'c_char_p', 'c_double',
            'c_float', 'c_int', 'c_int16', 'c_int32', 'c_int64', 'c_int8',
            'c_long', 'c_longdouble', 'c_longlong', 'c_short', 'c_size_t',
            'c_ssize_t', 'c_ubyte', 'c_uint', 'c_uint16', 'c_uint32',
            'c_uint64', 'c_uint8', 'c_ulong', 'c_ulonglong', 'c_ushort',
            'c_void_p', 'c_voidp', 'c_wchar', 'c_wchar_p',
            'ARRAY', 'BigEndianStructure', 'CDLL', 'CFUNCTYPE', 'DEFAULT_MODE',
            'LibraryLoader', 'LittleEndianStructure', 'POINTER', 'PYFUNCTYPE',
            'PyDLL', 'RTLD_GLOBAL', 'RTLD_LOCAL', 'SetPointerType',
            'addressof', 'byref', 'cast', 'pointer',
        ]
        for key in ctypes_members:
            self.known[key] = 'from ctypes import {}'.format(key)

    def _populate_existing_modnames(self, names):
        # Populates any name that corresponds to a top-level module
        from xdoctest import static_analysis as static
        for n in names:
            if n not in self.known:
                if static.modname_to_modpath(n) is not None:
                    self.known[n] = 'import {}'.format(n)


def autogen_imports(fpath=None, source=None, importable=None,
                    search_modnames=True):
    """
    Generate lines of code that would fix the undefined names.

    This will work out of the box for common aliases (e.g. np) and for names
    that correspond to top-level modules, but anything else will require
    external hints to be passed in via importable.

    Args:
        fpath (PathLike): path to the module in question

        source (str): source code of file (mutually exclusive with fpath)

        importable (dict | Importables): mapping from names to import lines

        search_modnames (bool): if True, searches PYTHONPATH for existing
            modnames that match undefined unknown names.

    Example:
        >>> from xinspect.autogen import *  # NOQA
        >>> import ubelt as ub
        >>> source = ub.codeblock(
        >>>     '''
        >>>     p = os.path.dirname(join('a', 'b'))
        >>>     p = os.path.dirname(join('a', 'b'))
        >>>     glob.glob(p)
        >>>     print(antigravity)
        >>>     ''')
        >>> # Generate a list of lines to fix the name errors
        >>> lines = autogen_imports(source=source)
        >>> print(lines)
        ['import antigravity', 'import glob', 'from os.path import join', 'import os']
        >>> # After fixing the errors, the file does not need modification
        >>> newsource = chr(10).join(lines) + chr(10) + source
        >>> newlines = autogen_imports(source=newsource)
        >>> print(newlines)
        []
    """

    if not (bool(source) ^ bool(fpath)):
        raise ValueError('Must specify exactly one fpath or source')

    # Search for undefined names in a module
    names = undefined_names(fpath, source=source)

    # Use predefined
    if importable is None:
        importable = Importables()
        importable._use_recommended_defaults()

    importable = Importables(importable)
    if search_modnames:
        importable._populate_existing_modnames(names)

    # Add any unregistered names if they correspond with a findable module
    have_names = set(importable.known).intersection(set(names))
    missing = set(names) - set(have_names)
    if missing:
        message = ('Warning: unknown modules {}'.format(missing))
        print(message)
        warnings.warn(message)

    import_lines = [importable.known[n] for n in sorted(have_names)]
    return import_lines
